// Qe.cpp : Defines the entry point for the console application.
//
/*

	Simple program for Quantum Circuit synthesis exhaustive search
	No documentation file for this program is provided
	al;l comments should be enough to understand it

  In the present configuration this program searches for circuit that are similar to the resulting ones
  however can be modified to remember all possible ones.



*/

/////////////////////////////////////////////////////////////////////////////
// The one and only application object


#include "stdafx.h"
#include "Qe.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//sub structure of Individ in order 
//to store more representation for 
//one circuit
struct qData
{
	int cost;			// cost of this gate/circuit -- not used
	string my_string;	// first representation
	qData *next;		// link to the nest possible representation

};
//structure representing quantum
//gates read from the input file
struct qGate
{
       int numIO;									// number of input/output wires
	   int Cost;									// cost of the gate
	   char representation;							// representation
       dComp gateMatrix1[MAXGATEDEM][MAXGATEDEM];	// matrix representing this gate
};

//1-st type of individual used 
//for exhaustive search without
//matching to a final gate
struct Individual
{
       int ioNumber;								// number of input/output wires
	   int cost;									// cost of the gate
       string my_string;							// representation
       dComp gateMatrix1[MAXGATEDEM][MAXGATEDEM];	// matrix representing this circuit


};

//2-nd type of individual storing
//the matrix of circuit, pointer
//to its representation and possible
//followers
struct Crc
{
       int ioNumber;								// number of input/output wires
	   int Counter;									// counter of all representation so far found
	   qData *Data;									// link to structures with representation
	   char representation;							// representation -- initial
       dComp gateMatrix1[MAXGATEDEM][MAXGATEDEM];	// matrix representing this circuit

};


// holds GA functions
class GA
{
public:
	GA(){}
  ~GA(){};
  dComp I;										//complex variable
  int numofgates;								// number of gates from the file
  int mumallcircuits;							// number or resulting circutis

  qGate *gateArray[MAXNUMOFGATES];				// an array of object qGate
  Crc *resultCircuits[MAXNUMBER];				// array of pointers to all generated circuits - not always used
  Individual *INDV;								// pointer to the only required individual
  ifstream in_stream;							// input file stream
  ofstream out_stream;							// output file stream

  void Initialize();							// initialize
  void individualMatrix(bool);					// calculate the matrix representation
  void closeStream();							// close the stream to files
  void finalOutput();							// outputs all at the end	
  void output();								// outputs something whener required			
  void output(Individual*);						// output the pointed individual
  bool setTerminatingCondition();				// set the finish
  Individual setIndv(Individual*);				// a copy method
  void setIndv(Individual*, Individual*);		// a copy method using pointers

  //for minimizer
  void synthSegment();							// synthesize a new segment of the circuit
  void compareCirc();							// compare a circuit to all available

  //criteron for search
  int Segments;									// number of segments - maximal
  int Counter; // number of permutations for all circuit generation based on the permutation of gates 
				// per segment multiplied by the number of segments
  int IO;										// number of input./output 
  int gateCounter[MAXSEGNUM];					// binary counter 
  int population;								// number of all synthesized gates


private:

  int generationCondition; // termination criterian

  qGate tensorProduct(qGate*, qGate*);	// tensor product of gate A and gate B
  qGate matrixProduct(qGate*, qGate*);	// matrix product of gate A and gate B
  int getGate(char);					// returns a gate pointed by a character
  void initializeGates();				// intialize all gates from the file


};



/////////////////////////////////////////////////////////////////////////////
// The one and only application object

CWinApp theApp;

using namespace std;

int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
	int nRetCode = 0;

	// initialize MFC and print and error on failure
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
	{
		cerr << _T("Fatal Error: MFC initialization failed") << endl;
		nRetCode = 1;
	}
	else
	{


		// ignition
		GA g;
		// ignition
		g.Initialize();
		int a = 0;
		// life cycle
		while (!(g.setTerminatingCondition()))
		{
			g.synthSegment();
			g.individualMatrix(false);
			g.compareCirc();
			a++;

		//---------printing---------------
		}
			g.finalOutput();
			g.closeStream();
	}
	return nRetCode;
}

/*
	
     calculation of individual fitness!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     and of individual matrix representation
	 setting the fitness for each individual also
	 because each individual stores only a representation 
	 string not the whole circuit

*/

void GA::individualMatrix(bool display)
{    //temporary storage values
     float Error = 0;
	 int Time = 0;
	 float Cost = 0;

     //three free gates for operations
	 //finalG is the mtrix represenation for fitness evaluation
     qGate *temp0, *temp1;
	 string S = (*INDV).my_string;

	 int j = 0;
     for (int a = 0;a<= numofgates;a++)
        if ((*gateArray[a]).representation == S.at(j))
		{
			temp0 = gateArray[a]; 
			Cost+= (*gateArray[a]).Cost; 
			break;
		}
	 j++;

	 while (j<=S.length()-1)
	 {
		 for ( a = 0;a<= numofgates;a++)
			 if ((*gateArray[a]).representation == S.at(j))
			 {
				 temp1 = gateArray[a]; 
				 Cost+= (*gateArray[a]).Cost; 
				 break;
			 }

			 //calculation of each segment
			 //recursive calculation of the circuit
		 temp0 = &matrixProduct(temp0, temp1);


     	 j++;
	 }


	 for ( a = 0; a<(int(pow(2,(*temp0).numIO)));a++)
	 {
		 for (int b = 0; b<(int(pow(2,(*temp0).numIO)));b++)
		 {
			 (*INDV).gateMatrix1[a][b] = (*temp0).gateMatrix1[a][b];
		 }
	 }
		
	if (display)
	{
		out_stream << " -------- " <<endl;
			for (int a = 0; a<(int(pow(2,(*temp0).numIO)));a++)
			{
				for (int b = 0; b<(int(pow(2,(*temp0).numIO)));b++)
					out_stream << (*temp0).gateMatrix1[a][b] <<" ";
					out_stream << endl;
			}
	}
}



void GA::output()
{
	qGate *qG;

     out_stream << "There are " << numofgates << " gates from the input file and synthesized.\n";
     for (int i = 0; i < numofgates; i++)
     {
		 qG = gateArray[i];

         out_stream << "The gate number " << i << " is: " << endl;
         out_stream << "And the gate representation is: " << (*gateArray[i]).representation << endl;
         out_stream << "Number of inputs is: " << (*qG).numIO << endl;

         for (int j = 0; j < (int(pow(2,(*qG).numIO))); j++)
         {
             for (int k = 0; k < (int(pow(2,(*qG).numIO))); k++)
             {
                 out_stream << (*qG).gateMatrix1[j][k] << " ";
             }
             out_stream << endl;
         }
     }
}

void GA::output(Individual *I)
{
     out_stream << "---------------------------------------------------------- "<<endl;
     out_stream << "---------------------------------------------------------- "<<endl;

     out_stream << "Found this representation similar to one of searched gates "<<endl;
     out_stream << "number of IO "<<(*I).ioNumber<<endl;
     out_stream << "representation "<<(*I).my_string<<endl;
     
	 for (int i = 0; i < (*I).ioNumber; i++)
     {
         for (int j = 0; j < (int(pow(2,(*I).ioNumber))); j++)
         {
             for (int k = 0; k < (int(pow(2,(*I).ioNumber))); k++)
             {
                 out_stream << (*I).gateMatrix1[j][k] << " ";
             }
             out_stream << endl;
         }
     }

     out_stream << "---------------------------------------------------------- "<<endl;
     out_stream << "---------------------------------------------------------- "<<endl;

}

void GA::finalOutput()
{
	Crc *I;
	qData *qD;
	int c;

     out_stream << "There are " << population << " gates synthesized.\n";
     for (int i = 0; i < population; i++)
     {
		 I = resultCircuits[i];

         out_stream << "The gate number is: " << i << endl;
         out_stream << "The gate IOs are" << (*I).ioNumber << endl;
         out_stream << "Number of inputs is: " << (*I).ioNumber << endl;


		 qD = I->Data;
		 c = 0;
		 while(c<(*I).Counter)
		 {
				out_stream << "Representation  "<<(qD->my_string)<<endl;
				qD = qD->next;
		 }


         for (int j = 0; j < (int(pow(2,(*I).ioNumber))); j++)
         {
             for (int k = 0; k < (int(pow(2,(*I).ioNumber))); k++)
             {
                 out_stream << (*I).gateMatrix1[j][k] << " ";
             }
             out_stream << endl;
         }
     }
}

// Restrictions:
// 1. no fan out - ex, 5 inputs has to have 5 outputs
// 2. maximum 5 I/O (geometric increase of complexity with more I/Os)
void GA::Initialize()
{
     char inFile[100];
     char outFile[100];

     // Open input and output files.
     cout << "Please input the source file" << endl;
     cin >> inFile;


     in_stream.open(inFile);
     if (in_stream.fail())
     {
        cout << "Input file opening failed.\n";
        exit(1);
     }

     cout << "\nPlease input the output file" << endl;
     cin >> outFile;

     out_stream.open(outFile);
     if (out_stream.fail())
     {
        cout << "Output file opening failed.\n";
        exit (1);
     }
     // initialize population also initialize quantum gates used
     // read from file
	 //populationNumber = 50;

	 population = 0;
	 INDV = new Individual;

	 generationCondition = 0;

     initializeGates();
	 
	 out_stream << "\nintialization done" << endl;


	 for (int a = 0;a<MAXSEGNUM;a++)
		 gateCounter[a] = 0;
 

}



int GA::getGate(char b)
{
	for (int a = 0;a<numofgates;a++)
		if ((*gateArray[a]).representation == b) return a;
  
}



bool GA::setTerminatingCondition()
{
	int result = 0;
	for (int a=0;a<Segments;a++)
		result += gateCounter[a];

	if (result >= (Segments*(numofgates-1)))
	{
		out_stream<<" generation max reached "<<endl<<endl<<endl;
		return true;
	}
	else {
		generationCondition++;
		return false;
	}
}


// or the understanding of the format of the input file see the attached 
// templatesand examples
void GA::initializeGates()
{
	I = dComp (0., 1.); //  Declare complex double variables
	double x;


     //--------------------------------------------------------------
     // WON'T BE USED FOR NOW
     // read first number from the file
	//first line in file should be indicating how many segments are allowed to be searched
	in_stream >> Segments;

	if (Segments >= MAXSEGNUM) Segments = MAXSEGNUM;

	//second line should indicate the number of wires of the circuit
	in_stream >>  IO;

	out_stream<< " output file of synthesis "<<endl;
	out_stream<< " max number of segments  "<< Segments<<endl;
	out_stream<< " number of IO  "<< IO <<endl;

	(*INDV).ioNumber = IO;

	//third line - the toal number of gates in the file
	//each gate should have 2 nubers before the matrix
    in_stream >> numofgates;
	out_stream<< " number of gates  "<< numofgates <<endl;
	out_stream<< " size of the array for gates  "<< MAXNUMOFGATES<<endl;

	
     // get first gate
    for (int a = 0; a < numofgates; a++)
     {

		 out_stream<< "gate number : " <<a<<endl;

		gateArray[a] = new qGate;

		//first the nuber of wires
         in_stream >> (*gateArray[a]).numIO;
         (*gateArray[a]).representation = char(a+65); // starting from A

		 // then the cost
		 in_stream >> (*gateArray[a]).Cost;

		 out_stream<< "gate representation: "<<(*gateArray[a]).representation <<endl;
		 out_stream<<"  Cost of the gate  "<< (*gateArray[a]).Cost <<endl;
		 out_stream<<"  IO of the gate  "<< (*gateArray[a]).numIO <<endl;

         for (int j = 0; j < (int(pow(2,(*gateArray[a]).numIO))); j++)
         {
             for (int k = 0; k < (int(pow(2,(*gateArray[a]).numIO))); k++)
             {
				in_stream >> x;
				out_stream<< x <<' ';
                 if (x == 2)
                  (*gateArray[a]).gateMatrix1[j][k] = I;
                 else if (x == -2)
                         (*gateArray[a]).gateMatrix1[j][k] = -I;
                      else if (x == 4)
							(*gateArray[a]).gateMatrix1[j][k] = exp(complex<double>(0, (3.1416/4)));
							else if (x == 5)
								(*gateArray[a]).gateMatrix1[j][k] = complex<double>(0.5, 0.5);
								else if (x == -5)
									(*gateArray[a]).gateMatrix1[j][k] = complex<double>(0.5, -0.5);
									else (*gateArray[a]).gateMatrix1[j][k] = x;
				out_stream<< (*gateArray[a]).gateMatrix1[j][k]<<' ';
             }
			out_stream<< endl;
         }
     }

      out_stream<< "------------------------------------------"<<endl;
      out_stream<< "------------------------------------------"<<endl;
      out_stream<< "------------------------------------------"<<endl;
      out_stream<< "------------------------------------------"<<endl;
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
//here follows reading final gates for searching.


	  string temo;
	  in_stream >>temo;
		out_stream<< temo <<endl;
  	  in_stream >>temo;
		out_stream<< temo <<endl;
	  in_stream >>temo;
		out_stream<< temo <<endl;

	  in_stream >> population;

     // get first gate
    for (a = 0; a < population; a++)
	{
		out_stream<< "gate number : " <<a<<endl;

		resultCircuits[a] = new Crc;
		resultCircuits[a]->Counter = 0;
         in_stream >> (*resultCircuits[a]).ioNumber;
         (*resultCircuits[a]).representation = char(a+65); // starting from A

		 out_stream<< "gate representation: "<<(*resultCircuits[a]).representation <<endl;
		 out_stream<<"  IO of the gate  "<< (*resultCircuits[a]).ioNumber <<endl;

         for (int j = 0; j < (int(pow(2,(*resultCircuits[a]).ioNumber))); j++)
         {
             for (int k = 0; k < (int(pow(2,(*resultCircuits[a]).ioNumber))); k++)
             {
				in_stream >> x;
				out_stream<< x <<' ';
                 if (x == 2)
                  (*resultCircuits[a]).gateMatrix1[j][k] = I;
                 else if (x == -2)
                         (*resultCircuits[a]).gateMatrix1[j][k] = -I;
                      else if (x == 4)
							(*resultCircuits[a]).gateMatrix1[j][k] = exp(complex<double>(0, (3.1416/4)));
							else if (x == 5)
								(*resultCircuits[a]).gateMatrix1[j][k] = complex<double>(0.5, 0.5);
								else if (x == -5)
									(*resultCircuits[a]).gateMatrix1[j][k] = complex<double>(0.5, -0.5);
									else (*resultCircuits[a]).gateMatrix1[j][k] = x;
				out_stream<< (*resultCircuits[a]).gateMatrix1[j][k]<<' ';
             }
			out_stream<< endl;
         }
     }
      out_stream<< "------------------------------------------"<<endl;
      out_stream<< "------------------------------------------"<<endl;
      out_stream<< "------------------------------------------"<<endl;

    
}


qGate GA::tensorProduct(qGate *A, qGate *B)
{

      qGate pGate;
      for (int i = 0; i < (int(pow(2,(*A).numIO))); i++)
      {    for (int j = 0; j < (int(pow(2,(*A).numIO))); j++)
           {   for (int k = 0; k < (int(pow(2,(*B).numIO))); k++)
               {   for (int l = 0; l < (int(pow(2,(*B).numIO))); l++)
                   {  
                      pGate.gateMatrix1[i*(int(pow(2,(*B).numIO)))+k][j*(int(pow(2,(*B).numIO)))+l] = (*A).gateMatrix1[i][j] * (*B).gateMatrix1[k][l];
                   }
              }

          }
      }


      pGate.numIO = ((*A).numIO + (*B).numIO);

      return pGate;
}


qGate GA::matrixProduct(qGate *A, qGate *B)
{ 
	qGate pGate;
       //rows of the first gate
      for (int i = (int(pow(2,(*A).numIO))-1);i>=0; i--)
      {   //colums of the second gate
          for (int j = 0; j < (int(pow(2,(*B).numIO))); j++)
          {   //colums of the first matrix
		int l = (int(pow(2,(*A).numIO))); 
              for (int k = 0; k < (int(pow(2,(*A).numIO))); k++)
              { //rows of the second
				pGate.gateMatrix1[j][i] = pGate.gateMatrix1[j][i] + (*A).gateMatrix1[k][i] * (*B).gateMatrix1[j][k];
 				l--;
              }
          }
      }
      pGate.numIO = (*A).numIO;
      return pGate;

}

void GA::closeStream()
{
     in_stream.close();
     out_stream.close();
}


void GA::compareCirc()
{
	int a, c;
	Individual *I = INDV;
	bool matrixTest;
	qData *qD;

	a = 0;
	while (a<population)
	{

		matrixTest = true;

			cout<<" comparing to the "<<a<<" gate "<<endl;

		for (int i = 0; i < (int(pow(2,(*I).ioNumber))); i++)
		{
			for (int j = 0; j < (int(pow(2,(*I).ioNumber))); j++)
			{
				if ((*I).gateMatrix1[i][j] != (*resultCircuits[a]).gateMatrix1[i][j])
					  matrixTest = false;

			}

		}


		if (matrixTest) 
		{
			c = 0;
			qD = resultCircuits[a]->Data;
			for(int b =0;b<resultCircuits[a]->Counter;b++)
					qD = qD->next;
			
			qD = new qData;

			qD->my_string = I->my_string;

			out_stream<<qD->my_string<<endl;

			for (int i = 0; i < (int(pow(2,(*I).ioNumber))); i++)
			{
				for (int j = 0; j < (int(pow(2,(*I).ioNumber))); j++)
					out_stream<<(*resultCircuits[a]).gateMatrix1[i][j];
	
				out_stream<<endl;
			}
			out_stream<<endl;
			out_stream<<endl;
			

			break;
		}
		else a++;
	} 
	
}


Individual GA::setIndv(Individual *indv2)
{
	//Method simulatin an operator for structures Individual
	//from  individual2 to individual1
		Individual indv1;
		indv1.ioNumber = indv2->ioNumber;
		indv1.cost = indv2->cost;

		indv1.my_string = indv2->my_string;
		for (int i = 0; i < (int(pow(2,(*indv2).ioNumber))); i++)
		    for (int j = 0; j < (int(pow(2,(*indv2).ioNumber))); j++)
                      indv1.gateMatrix1[i][j] = (*indv2).gateMatrix1[i][j];
		return indv1;

}
void GA::setIndv(Individual *indv1, Individual *indv2)
{
	//Method simulatin an operator for structures Individual
	//from  individual2 to individual1
		indv1->ioNumber = indv2->ioNumber;
		indv1->cost = indv2->cost;
		indv1->my_string = indv2->my_string;

		for (int i = 0; i < (int(pow(2,(*indv2).ioNumber))); i++)
		    for (int j = 0; j < (int(pow(2,(*indv2).ioNumber))); j++)
                      indv1->gateMatrix1[i][j] = indv2->gateMatrix1[i][j];

}


void GA::synthSegment()
{
	int carry, a;
	INDV->my_string = "";

	if (gateCounter[0] < numofgates-1)
	{
		gateCounter[0]++;
	}
	else
	{
		gateCounter[0] = 0;
		carry = 1;
		a = 1; 
		while(true)
		{
			if 	(gateCounter[a] == numofgates-1)
			{
				carry = 1;
				gateCounter[a] = 0;
				
			}
			else
			{
				gateCounter[a]++;
				carry = 0;
				break;
			}

			if (a < numofgates-1) a++;
			else break;
		
		}
	}

	carry = 0;
	for (a = Segments-1;a>=0;a--)
	{
		if (carry == 0 && gateCounter[a] != 0)
			carry = 1;

		if (carry != 0)
			INDV->my_string.insert(0, (*gateArray[gateCounter[a]]).representation);

	}

}
